
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <SDL/SDL.h>
#include <stdio.h>
#include <string.h>

#include "screen.h"
#include "emu.h"
#include "video.h"
#include "conf.h"

#include "blitter.h"
#include "effect.h"

blitter_func blitter[] = {
    {"soft","Software blitter",blitter_soft_init, NULL, blitter_soft_update, blitter_soft_fullscreen,
     blitter_soft_close},
    {"opengl","Opengl blitter",blitter_opengl_init, blitter_opengl_resize, blitter_opengl_update,
     blitter_opengl_fullscreen, blitter_opengl_close},
    {"yuv","YUV blitter (YV12)",blitter_overlay_init, blitter_overlay_resize, blitter_overlay_update,
     blitter_overlay_fullscreen, blitter_overlay_close},
    {NULL,NULL,NULL,NULL,NULL,NULL,NULL}
};

effect_func effect[] = {
    {"none","No effect",1,1,NULL, NULL},
    {"scanline","Scanline effect",2,2,effect_scanline_init, effect_scanline_update},	// 1
    {"scanline50","Scanline 50% effect",2,2,effect_scanline_init, effect_scanline50_update},	// 2
    {"scale2x","Scale2x effect",2,2,effect_scale2x_init, effect_scale2x_update},	// 3
    {"scale3x","Scale3x effect",3,3,effect_scale3x_init, effect_scale3x_update},	// 3
    {"scale4x","Scale4x effect",4,4,effect_scale4x_init, effect_scale4x_update},	// 3
    {"scale2x50","Scale2x effect with 50% scanline",2,2,effect_scale2x_init, effect_scale2x50_update},	// 4
    {"scale2x75","Scale2x effect with 75% scanline",2,2,effect_scale2x_init, effect_scale2x75_update},	// 5
    {"doublex","Double the x resolution (soft blitter only)",2,1,effect_scanline_init, effect_doublex_update},	//6
#ifdef I386_ASM
    {"sai","SAI effect",2,2,effect_sai_init, effect_sai_update},	//7
    {"supersai","SuperSAI effect",2,2,effect_sai_init, effect_supersai_update},	//8
    {"eagle","Eagle effect",2,2,effect_sai_init, effect_eagle_update},	//9
#endif
    {NULL,NULL,0,0,NULL,NULL}
};

/* Interpolation */
static SDL_Surface *tmp, *blend;

#ifdef I386_ASM
unsigned int scan_mask = 0xf7def7de;
void do_inner_interpolation_i386(Uint16 * dst, Uint16 * src);
#endif

void print_blitter_list(void) {
    int i=0;
    while (blitter[i].name!=NULL) {
	printf("%-12s : %s\n",blitter[i].name,blitter[i].desc);
	i++;
    }
}
void print_effect_list(void) {
    int i=0;
    while (effect[i].name!=NULL) {
	printf("%-12s : %s\n",effect[i].name,effect[i].desc);
	i++;
    }
}


Uint8 get_effect_by_name(char *name) {
    int i=0;
    
    while(effect[i].name!=NULL) {
	if (!strcmp(effect[i].name,name)) {
	    return i;
	}
	i++;
    }
    /* invalid effect */
    printf("Invalid effect.\n");
    return 0;
}

Uint8 get_blitter_by_name(char *name) {
    int i=0;

    while(blitter[i].name!=NULL) {
	if (!strcmp(blitter[i].name,name)) {
	    return i;
	}
	i++;
    }
    /* invalid blitter */
    printf("Invalid blitter, using soft blitter.\n");
    return 0;
}


SDL_bool screen_init()
{
    CONF_ITEM *cf_blitter,*cf_effect,*cf_interpol,*cf_scale,*cf_fs;

    /* screen configuration init */
    cf_blitter=cf_get_item_by_name("blitter");
    cf_effect=cf_get_item_by_name("effect");
    cf_interpol=cf_get_item_by_name("interpolation");
    cf_scale=cf_get_item_by_name("scale");
    cf_fs=cf_get_item_by_name("fullscreen");

    /* Initialization of some variables */
/*
    interpolation = conf.interpolation;
    nblitter = conf.nblitter;
    neffect = conf.neffect;
*/
    interpolation=CF_BOOL(cf_interpol);
    nblitter=get_blitter_by_name(CF_STR(cf_blitter));
    neffect=get_effect_by_name(CF_STR(cf_effect));
    fullscreen=CF_BOOL(cf_fs);

    if (CF_VAL(cf_scale) == 0)
	scale = 1;
    else
	scale = CF_VAL(cf_scale);

    /* Init of video blitter */
    if ((*blitter[nblitter].init) () == SDL_FALSE)
	return SDL_FALSE;

    /* Init of effect */
    if (neffect > 0)
	if ((*effect[neffect].init) () == SDL_FALSE)
	    return SDL_FALSE;

    /* Interpolation surface */
    blend = SDL_CreateRGBSurface(SDL_SWSURFACE/*(conf.hw_surface ? SDL_HWSURFACE : SDL_SWSURFACE)*/,
				 336, 256, 16, 0xF800, 0x7E0, 0x1F, 0);

    return SDL_TRUE;
}

SDL_bool screen_resize(int w, int h)
{
    //nblitter = conf.nblitter;
    if ((*blitter[nblitter].resize) (w, h) == SDL_FALSE)
	return SDL_FALSE;
    return SDL_TRUE;
}
static inline void do_interpolation()
{
    Uint16 *dst = (Uint16 *) blend->pixels + 16 + (336 << 4);
    Uint16 *src = (Uint16 *) buffer->pixels + 16 + (336 << 4);
    Uint32 s, d;
    Uint8 w, h;
#ifdef I386_ASM
    do_inner_interpolation_i386(dst, src);
#else
    /* we copy pixels from buffer surface to blend surface */
    for (w = 224; w > 0; w--) {
	for (h = 152; h > 0; h--) {
	    s = *(Uint32 *) src;
	    d = *(Uint32 *) dst;

	    *(Uint32 *) dst =
		((d & 0xf7def7de) >> 1) + ((s & 0xf7def7de) >> 1) +
		(s & d & 0x08210821);

	    dst += 2;
	    src += 2;
	}
	src += 32;
	dst += 32;
    }
#endif
    /* Swap Buffers */
    tmp = blend;
    blend = buffer;
    buffer = tmp;
}

void screen_update()
{
    if (interpolation == 1)
	do_interpolation();

    if (neffect != 0)
	(*effect[neffect].update) ();

    (*blitter[nblitter].update) ();
}

void screen_close()
{
    SDL_FreeSurface(blend);
}

void screen_fullscreen()
{
    blitter[nblitter].fullscreen();
}
